
#include <iostream>
#include <assert.h>
#ifdef MMN_isimn
  #include "gpibn/gp_ifn.hxx"
#else
  #include "gpib/gp_if.hxx"
#endif
#include "gpibc/gpib.h"
#include "gpibc/gumi_msgs.hxx"



  namespace Gif=INdepGpib100;
  #define GAD_ISIM 3		/* simulator's GPIB address	*/

  
  /*
   * Standard Event Register class
   * 11.4.2.2-3
   */
  class SerReg {
    int _fEv;	/* event  	*/
    int _fEn;   /* enable 	*/
  public:

    bool sum() { return (_fEv & _fEn) != 0; }
    int read() { int f=_fEv; _fEv=0; return f; }
    SerReg &setbit( int f) { _fEv |= f; return *this; }
    SerReg &setEnable( int f) { _fEn |= f; return *this; }
    int fEnable() { return _fEn; }
    SerReg():_fEv(0), _fEn(0) {}
    ~SerReg() {}
  };

  class Instrument:public Gif::Instrument {
    std::ostream &_os;		/* diagnostic log stream */
    std::ostream &os() { return _os; }
    SerReg _serSes;		/* Standard Event Status 	*/
    SerReg _serSrq;		/* Service Request Enable	*/
    SerReg &serSes() { return _serSes; }
    SerReg &serSrq() { return _serSrq; }

    Failable &updateStatus(Gif::GpibDevice *pgd);
   public:
    /* 
     * must return after command execution 
     * or initialization
     */
    int parse(Gif::GpibDevice *pgd);

    Instrument(std::ostream &os):_os(os) {}
    virtual ~Instrument() {}
  };

  Failable& Instrument::updateStatus(Gif::GpibDevice *pgd) {
    int fStb(serSes().sum() ? GPIB_4882_STB_ESB  : 0);
    return pgd->setStatus( fStb, serSrq().fEnable() );
  }

  /*
   * instrument parser
   */
  int Instrument::parse(Gif::GpibDevice *pgd) {
    std::stringbuf sbCmd;

    int fEND(false);

    do {
      /* fill the buffer */
      while ((pgd->sgetc()) != EOF) sbCmd.sputc(pgd->sbumpc());

      switch (pgd->getEof()) {
        /* 
	 * status update
	 */
       case Gif::M_STATUS:
	os() << ">status 0x" << std::hex << pgd->fStat() << std::endl;
	break;

        /*
	 * End of message received from controller 
	 */
       case Gif::M_END:
	os() << ">end" << std::endl;
	fEND=true;
	break;

        /*
	 * GET trigger received
	 */
       case Gif::M_GET:
	os() << ">trigger" << std::endl;
	break;

       /* 
        * events loosing the buffer 
	*/
       case Gif::M_FAIL:
	os() << ">failed" << std::endl;
	return -1;
       case Gif::M_INITIALIZE:
	/* 6.3.2.1 */
	os() << ">initialize" << std::endl;
	return 0;			/* MUST return */
       case Gif::M_UNTERMINATED:
	/* 6.3.2.2 */
	os() << ">unterminated" << std::endl;
	if (serSes().setbit(GPIB_4882_ESR_QYE).sum() && !updateStatus(pgd)) return -1;
	return 0;
       case Gif::M_INTERRUPTED:
	/* 6.3.2.3 */
	os() << ">interrupted" << std::endl;
	if (serSes().setbit(GPIB_4882_ESR_QYE).sum() && !updateStatus(pgd)) return -1;
	return 0;
       default:
	assert(0);
      }
    } while (!fEND);

    /* execute program message */
    std::string sgCmd(sbCmd.str());


    if (sgCmd == "POFF\n") {
      /* last command executed */
      os() << "power off, bye..." << std::endl;
      return -1;
    }
    if (sgCmd == "STEN\n") {
      os() << "set status" << std::endl;

      /* enable the ESB bit in the status byte */
      serSrq().setEnable(GPIB_4882_STB_ESB);

      /* set the ESR Query Error enable and status bits */
      if (serSes().setEnable(GPIB_4882_ESR_QYE).setbit(GPIB_4882_ESR_QYE).sum() && !updateStatus(pgd)) return -1;
    }
    else if (sgCmd == "*ESR?\n") {
      std::ostream osData(pgd);
      pgd->setEndNL();
      os() << "send serSes()" << std::endl;
      osData << serSes().read() << std::endl;
    }
    else if (sgCmd == "SEND?") {
      /* send sample binary data */
      os() << "send start" << std::endl;

      pgd->clearEnd();
      for (int i=0; i<345; i++) {
	if (i==344) pgd->setEndNext();
	pgd->sputc(i);
      }
      pgd->pubsync();

      os() << "send done" << std::endl;
    }

    /*
     * parse() MUST return after each message parsed
     */
    return 0;		
  }


  int main(int argc, char **argv) {

    if (argc != 2) {
      std::cerr << "usage: " << argv[0] << " <gpib-interface>" << std::endl;
      return 2;
    }

    /*
     * construct the GPIB device interface 
     */
  #ifdef MMN_gputn	// network support available
    Gif::GpibInterfaceDeviceNet gd(argv[1]);
  #else
    Gif::GpibInterfaceDevice gd(argv[1]);
  #endif

    if (!gd.isGood()) {
      std::cerr << "init failure:" << gd << std::endl;
      return 2;
    }
    Instrument in(std::cerr);
    Gif::GpibDevice *pgd(gd.pGud());

    if (!pgd->open(0, GAD_ISIM, false, 0)) {
      std::cerr << "open failed:" << *pgd << std::endl;
      return 2;
    }

    if (!pgd->run(&in)) {
      std::cerr << "run failed:" << *pgd << std::endl;
      return 2;
    }

    if (!pgd->close()) {
      std::cerr << "close failed:" << *pgd << std::endl;
      return 2;
    }
    return 0;
  }

