
/*
 * libmsg - raw GPIB message base library
 * #LEGAL#
 */
#ifndef _GISI_GISU_HXX_
#define _GISI_GISU_HXX_

#include "cplRaw/sm.hxx"
#include "cplRaw/dbg.hxx"
#include "gpib/gumi_user.hxx"
#include "gpibc/gumi_msgs.hxx"
#include <iostream>
#include <vector>

class Block;

namespace INdepGpib100 {
  namespace Umi=INdepUmi100;
  namespace Gif=INdepGpib100;

  typedef int DelE;

  /** A GpibAddress represents a GPIB bus address. A GPIB address has a primary address part
   *  and on optional secondary part.
   */
  class GpibAddress {
    int _iPa;
    bool _fSa;
    int _iSa;
   public:
    /** 
     *  Comparison operator. Two GpibAddress objects are equal if they represent
     *  the same primary and secondary (optional) GPIB address.
     */
    friend bool operator==(const GpibAddress &gaL, const GpibAddress& gaR) {
      return (
        gaL._iPa == gaR._iPa &&
        gaL._fSa == gaR._fSa &&
        gaL._iSa == gaR._iSa 
      );
    }
    friend bool operator!=(const GpibAddress &gaL, const GpibAddress& gaR) { return !operator==(gaL, gaR); }
    bool fSa() const { return _fSa; } 		/**< secondary address present */
    int iPa() const { return _iPa; }		/**< primary address	*/
    int iSa() const { return _iSa; }		/**< secondary address, valid if fSa()  */

    /** GpibAddress constructor. Represents the GPIB address \a iPa and the
     *  optional secondary address \a iSa if \a fSa is true.
     */
    GpibAddress(int iPa, bool fSa=false, int iSa=0): _iPa(iPa), _fSa(fSa), _iSa(iSa) {}

    /** Default Constructor, equivalent to GpibAddress(0).
     */
    GpibAddress(): _iPa(0), _fSa(false), _iSa(0) {}
    ~GpibAddress() {}
  };

  /** An GpibAddrList is a vector of GpibAddress objects.
   */
  class GpibAddrList:public std::vector<GpibAddress> { };
  class GpibStbList:public std::vector<char> { };
  class GpibResList:public std::vector<short> { };

  /** GpibCmds is a list of commands to be send over the GPIB.
   *  It defines several members to construct command sequences easily.
   */

  class GpibCmds {
    enum {
      MAX_CMD=40		/**< Limit of commands to be sent	*/
    };
    char _mpch[MAX_CMD];
    char *_pch;
    GpibAddress _ga;	/* controller address	*/

    bool canput(unsigned i) { return _pch+i < _mpch+MAX_CMD; } 
    void put(int ch) { *(_pch++)=ch; }

   public:

    void setupReceive( const GpibAddress &gaFrom);	/**< Setup to receive from \a gaFrom. inserts UNL MLA [MSA] TAG [SCG] */
    void setupReceive( );				/**< Setup to receive from unspecified talker. Inserts UNL MLA [MSA] */
    void setupSend( const GpibAddress& gaTo );
    void setupSendAdd( const GpibAddress &gaTo );
    void setupSend( const GpibAddrList &gal );
    void cmdAdd( int i);
    void addr( int cmd, const GpibAddress &ga);

    unsigned size() { return _pch-_mpch; }		/**< count of commands buffered so far	*/
    char *mpch() { return _mpch; }    
    char *pch() { return _mpch; }    			/**< buffer to send commands from	*/
    void clear() { _pch=_mpch; }			/**< clear and reuse buffer		*/
    GpibCmds(const GpibAddress &ga):_pch(_mpch), _ga(ga) {} /**< Construct a GpibCmds. \a ga if the controllers (i.e. our) address */
    ~GpibCmds() {}
  };

  /**
   * GPIB user base class (Controller)
   *
   * provides imperative functions for conducting the bus activity.
   *
   * Errors:
   *
   *  Any error in communicating with one of the UMI driver channels
   * results in a BAD condition, i.e. this object is unusable (isBad()).
   *
   *  Errors returned by the UMI driver channel are related to the
   * failed transaction only, i.e. another transaction may be started.
   * (isFail()). 
   * 
   */
  class GpibControl:public SmM, public DebugLog {
    Umi::Umi *_pum;
    int _iTrans;

    ErrE _eErr;				/* result of last call		*/
    Gif::StatE _fStat;			/* stat of last successful call	*/
    GpibInfo _gi;			/* driver capabilities		*/
    GpibAddress _ga;			/* setup by open()		*/
    int _iPaLast;			/* last foreign primaddr sent	*/
    bool _fSc;				/* system control ?		*/
    enum { S_CLOSED, S_OPEN };

    /* messages for GPIB interface */

    Umi::Umi *pum() const { return _pum; }
    void wait();
    void signal();

    void setTransNext() { _iTrans ++ ; }
    int iTransCur() { return _iTrans; }			/* current transaction id	*/

    /*
     * result managers
     */

    int setFailed( Umi::Umi *pum, int i ) {  return i; }	/* UMI failed		*/
    GpibControl &setFailed( Umi::Umi *pum ) {  return *this; }	/* UMI failed		*/

    GpibControl &setState( GpibResp *pgh ) { 
      _eErr=(ErrE)pgh->eErr(); 
      _fStat=(StatE)pgh->fStat(); 
      return *this; 
    }

    GpibControl &setState( GpibResp *pgh, int eState ) { 
      _fStat=(StatE)pgh->fStat(); 
      if (!(_eErr=(ErrE)pgh->eErr())) SmM::setState(eState); 
      return *this; 
    }

    //int setFailed( GpibResp *pgh, int i ) { _eErr=(ErrE)pgh->eErr(); return i; }
    //int setOk(int i) { _eErr=INdepGpib100::E_OK; return i; }
    GpibControl &setFailed(ErrE i) { _eErr=i; return *this; }

    void pushReceivedMessages(Umi::UmiControl*, Umi::UmiData*);	/* save the indications and continue */


    int cwr(Block &bl, const Umi::Time &ti, const GpibTermOut *gto, GpibTrf *ptr);

    GpibControl &getLines(unsigned *pf, const Umi::Time &tiSettle);

   public:
    friend std::ostream &operator <<( std::ostream &os, const GpibControl &lr );

    /*
     * state semantics
     * failed state is set by member operations 
     */
    ErrE eErr() const { return _eErr; }
    bool isErr(ErrE e) const { return _eErr == e; }

    bool isGood() const { return _eErr==0 && pum()->isGood(); }		/* fine 		*/
    bool isFail() const { return _eErr || !pum()->isGood(); }		/* bad request	 	*/
    bool isBad() const { return !pum()->isGood(); }			/* bad transport 	*/
    bool operator!() const { return isFail(); }


    int init(int iAddr);
    int fini();

    /* 488.2 11.3.3.4.1 conformant stb generation */
    GpibControl &updateStb(const GpibSta &gst);


    /* sync user fx's */
    GpibControl &open(bool fSc, const GpibAddress &);
    GpibControl &close();
    int isOpen() { return isState(S_OPEN); }

    GpibControl &ifc(const Umi::Time &ti=Umi::Time(0, 100000));
    GpibControl &ren(bool f);
    GpibControl &cac(bool f);	
    GpibControl &gts();			/* goto standby */
    GpibControl &ist(bool f);		/* individual status set */

    GpibControl &iblinec( unsigned *pCap );	/* capability to sense lines 	*/
    GpibControl &iblines( unsigned *pStat );

    GpibControl &rtl();			/* return to local */
    GpibControl &ppc(int v);		/* local pp config	 */
    GpibControl &rpp(char *chPpr, const Umi::Time &ti);		/* ppoll */
    //GpibControl &rsv(int);		/* rsv & stb */

    /*
     * all transfer functions return the count actually transfered
     * (even on failure)
     */
    int cmd( const char *pch, int c, const Umi::Time &ti );
    int cwr( const char *pch, int c, const Umi::Time &ti, const GpibTermOut & );
    int cwr( const char *pch1, int c1, const char *pch2, int c2, const Umi::Time &ti, const GpibTermOut & );

    /*
     * conducted read
     *
     * - read at most c bytes
     * - delimited according to gti
     * - timeout defined by ti, failed on timeout
     * - returns byte count read
     * - (*peDel) ' ' = not delimited, 'e' = eos (~EOI) delimited, 'E' = EOI delimited
     * - if pch is zero, no data is stored (shadow handshake)
     *
     * If a device read has been interrupted by a E_CLR failure,
     * E_DCAS is cleared and returned in the status word.
     */
    int crd( char *pch, int c, const Umi::Time &ti, const GpibTermIn &gti, DelE *peDel=0 );

    /*
     * wait for any specified flag or event active, failes on timeout
     *
     * For any unmasked flag, wait() returns when the flag is set.
     *
     * For any unmasked event, wait() returns when at least one false->true
     * transition has been detected since the most recent call to wait() with
     * the event unmasked.
     *
     * If no mask bit is set, wait() returns after ti has elapsed, 
     * updating the current status word. 
     *
     * if fIgnTmo is set, the E_TMO is not set on timeout.
     *
     */
    GpibControl &wait(Gif::StatE fStat, const Umi::Time &ti, bool fIgnTmo=false);	

    /*
     * Any successful operation updates the status flags
     */
    Gif::StatE fStat() { return _fStat; }

    /*
     * sev - synchronize with events
     *
     * returns the interface status and allows to wait for
     * specified events
     *
     */

    /* status */
    bool isSc() { return _fSc; }
    int iPa() { return _ga.iPa(); }
    int iSa() { return _ga.iSa(); }
    bool fSa() { return _ga.fSa(); }
    const GpibAddress &ga() const { return _ga; }
    int iPaLast() { return _iPaLast; }
    void clrPaLast(int i) { _iPaLast=i; }

    /*
     * standard commands
     */

    /* IEEE488.2 - 16.2.1 - SEND COMMAND */
    GpibControl &sendCmds(const char *, unsigned *pc, const Umi::Time &ti=Umi::Time(10));

    /* IEEE488.2 - 16.2.2 - SEND SETUP */
    GpibControl &sendSetup(const GpibAddrList &al, const Umi::Time&ti=Umi::Time(10));

    /* IEEE488.2 - 16.2.3 - SEND DATA BYTES */
    GpibControl &sendDataBytes(
      const char *, unsigned *pc, 
      const char *szTerm="\n",
      const GpibTermOut&gto=GpibTermOut().fEnd(true),
      const Umi::Time&ti=Umi::Time(10)
    );

    /* IEEE488.2 - 16.2.4 - SEND */
    GpibControl &send(
      const GpibAddress &, 
      const char *, unsigned *pc, 
      const char *szTerm="\n",
      const GpibTermOut&gto=GpibTermOut().fEnd(true),
      const Umi::Time&ti=Umi::Time(10)
    );

    GpibControl &send(
      const GpibAddrList &, 
      const char *, unsigned *pc, 
      const char *szTerm="\n",
      const GpibTermOut&gto=GpibTermOut().fEnd(true),
      const Umi::Time&ti=Umi::Time(10)
    );

    GpibControl &sendsz(
      const GpibAddress &ga, 
      const char *pch, 
      const char *szTerm="\n",
      const GpibTermOut&gto=GpibTermOut().fEnd(true),
      const Umi::Time&ti=Umi::Time(10)
    ) { 
      unsigned c(strlen(pch));
      return send(ga, pch, &c, szTerm, gto, ti); 
    }

    /* IEEE488.2 - 16.2.5 - RECEIVE SETUP */
    GpibControl &receiveSetup(const GpibAddress &ga, const Umi::Time&ti=Umi::Time(10));

    /* IEEE488.2 - 16.2.6 - RECEIVE RESPONSE MESSAGE */
    GpibControl &rcvRespMsg(char *, unsigned *pc, const GpibTermIn&gti=GpibTermIn(), const Umi::Time&ti=Umi::Time(10), DelE *peDel=0);


    /* IEEE488.2 - 16.2.7 - RECEIVE */
    GpibControl &receive(char *, unsigned *pc, const GpibAddress &, 
      const GpibTermIn&gti=GpibTermIn(),
      const Umi::Time&ti=Umi::Time(10),
      DelE *peDel=0
    );

    /* IEEE488.2 - 16.2.8 - SEND IFC */
    GpibControl &sendIfc(); 

    /* IEEE488.2 - 16.2.9.1 - DEVICE CLEAR SELECTED */
    GpibControl &devClear( const GpibAddrList &al, const Umi::Time&ti=Umi::Time(10));
    GpibControl &clear(const GpibAddress &gad, const Umi::Time &ti=Umi::Time(10));

    /* IEEE488.2 - 16.2.9.2 - DEVICE CLEAR ALL */
    GpibControl &devClear( const Umi::Time&ti=Umi::Time(10));

    /* IEEE488.2 - 16.2.10 - ENABLE LOCAL CONTROLS */
    GpibControl &enableLocal( const GpibAddrList &al, const Umi::Time&ti=Umi::Time(10));
    GpibControl &enableLocal( );

    /* IEEE488.2 - 16.2.11 - ENABLE REMOTE */
    GpibControl &enableRemote( const GpibAddrList &al, const Umi::Time&ti=Umi::Time(10));

    /* IEEE488.2 - 16.2.12 - SET RWLS */
    GpibControl &setRWLS( const GpibAddrList &al, const Umi::Time&ti=Umi::Time(10));

    /* IEEE488.2 - 16.2.13 - SEND LLO */
    GpibControl &sendLLO( const Umi::Time&ti=Umi::Time(10));

    /* IEEE488.2 - 16.2.14 - PASS CONTROL */
    GpibControl &passControl( const GpibAddress &al, const Umi::Time&ti=Umi::Time(10));

    /* IEEE488.2 - 16.2.15 - PPOLL */
    GpibControl &ppoll( unsigned *puRes,  const Umi::Time&ti=Umi::Time(0));

    /* IEEE488.2 - 16.2.16 - PARALLEL POLL CONFIGURE */
    GpibControl &ppollConfig( const GpibAddress &ga, int iLine, int iSense, const Umi::Time&ti=Umi::Time(10));

    /* IEEE488.2 - 16.2.17 - PARALLEL POLL UNCONFIGURE */
    GpibControl &ppollUnconfig( const GpibAddrList &ga, const Umi::Time&ti=Umi::Time(10));
    GpibControl &ppollUnconfig( const Umi::Time&ti=Umi::Time(10));

    /* IEEE488.2 - 16.2.18 - READ STATUS BYTE */
    GpibControl &GpibControl::readStatusByte(char *pch, const GpibAddress &gad, const Umi::Time &ti=Umi::Time(10));

    /* IEEE488.2 - 16.2.19 - TRIGGER */
    GpibControl &trigger(const GpibAddrList &ga, const Umi::Time &ti=Umi::Time(10));

    /* IEEE488.2 - 17.1 - RESET */
    GpibControl &reset(const GpibAddrList &ga, const Umi::Time &ti=Umi::Time(10));

    /* IEEE488.2 - 17.2 - FINDRQS */
    GpibControl &findRqs(char *pch, GpibAddrList::const_iterator *pit, const GpibAddrList &ga, const Umi::Time &ti=Umi::Time(10));


    /* IEEE488.2 - 17.3 - ALLSPOLL */
    GpibControl &allSpoll(GpibStbList *pgs, const GpibAddrList &ga, const Umi::Time &ti=Umi::Time(10));

    /* IEEE488.2 - 17.6 - FINDLSTN */
    GpibControl &findLstn(GpibAddrList *pga, const GpibAddrList &ga, const Umi::Time &tiSettle=Umi::Time(0,1000000), const Umi::Time &ti=Umi::Time(10));

    /* IEEE488.2 - 17.8 - TESTSYS */
    GpibControl &testSys(GpibResList *prl, const GpibAddrList &ga, const Umi::Time &ti=Umi::Time(10));

    GpibControl( Failable *pfa, const Umi::Factory *puf, int fDebug=0);
    ~GpibControl() ;
  };

}

#endif
