#ifndef _UMI_UMI_DRIVER_HXX
#define _UMI_UMI_DRIVER_HXX

/*
 * Universal Message Interface - Driver C++ part
 */

#include "umi.hxx"
#include "cplRaw/list.hxx"


namespace INdepUmi100 {

  /*
   * kernel message reps
   */
  class UmiMsgDown:public Link {
    UmiControl *_puc;
    UmiGetBuf *_pug;
   public:
    const UmiControl *puc() const { return _puc; }
    UmiGetBuf *pug() const { return _pug; }
    void init(UmiControl *puc, UmiGetBuf *pug) { _puc=puc; _pug=pug; }
    UmiMsgDown():_puc(0), _pug(0) {}
    ~UmiMsgDown() {}
  };

  class UmiMsgUp:public Link {
    UmiControl *_puc;
    UmiPutBuf *_pup;
   public:
    UmiControl *puc() const { return _puc; }
    UmiPutBuf *pup() const { return _pup; }

    void init(UmiControl *puc, UmiPutBuf *pup) { _puc=puc; _pup=pup; }
    UmiMsgUp():_puc(0), _pup(0) {}
    ~UmiMsgUp() {}
  };

  typedef const UmiMsgDown *Pmd;
  typedef const UmiMsgUp *Pmu;


  /*
   * self-contained message copies
   */
  class UmiMsgDownBuf:public UmiMsgDown, public UmiGetBuf, public UmiControlBuf { 
    char _mpchData[ SZ_DATA_MAX ];
   public:
    #ifndef _MM_NO_IOSTREAM
     friend std::ostream &operator<<(std::ostream &os, UmiMsgDownBuf &ud);
    #endif

    /*
     * copy in user messages
     */
    UmiControl& operator=(const UmiControl &uc) { return UmiControlBuf::operator=(uc); }

    UmiMsgDownBuf() ;
    ~UmiMsgDownBuf() ;
  };

  class UmiMsgUpBuf:public UmiMsgUp, public UmiPutBuf, public UmiControlBuf {
    char _mpchData[ SZ_DATA_MAX ];
   public:
    #ifndef _MM_NO_IOSTREAM
     friend std::ostream &operator<<(std::ostream &os, UmiMsgUpBuf &ud);
    #endif

    UmiMsgUpBuf();
    ~UmiMsgUpBuf();
  };

  /*
   * user/driver copy
   */
  void copy(UmiData *, const UmiMsgUpBuf*);
  void copy(UmiMsgDownBuf*, const UmiData*);

  /*
   * upper message handler interface definition
   */

  class UmiEnv {
   public:

    /*
     *  tiNext() schedules a call to UmiDriver::timer()
     *  not earlier than the "current time" + tiDelta (the target time).
     *
     *  It returns the target time, which must be
     *  used by timer() to qualify the time at call as
     *  being equal or later than than the target time.
     *
     *  tiDelta is limited to tiLim(). 
     *  
     */
    virtual Time tiNext( const Time &tiDelta )=0;		

    virtual void sigGet( int, void * );		
    virtual void sigPut( int, void * );		
    UmiEnv() {}
    virtual ~UmiEnv() {}
  };

  class UmiDriver;
  class UmiChannel : public Link {
    /* driver */
    UmiDriver*  _pud;
    UmiEnv*	_pue;

    /* channel number, identifies channel @ driver */
    int	        _iChan;		/* chan @ _pud			*/
    void*	_phEnv;		/* handle at _pue		*/

    /*
     * message pools
     */
    ChainRaw _cnUmuPool;
    ChainRaw _cnUmdPool;

    /*
     * downstream messages
     */
    ChainRaw _cnUmdQueue; 			/* messages queued for free channels 	*/
    ChainRaw _cnUmdDriver;			/* per channel message (single)		*/
    ChainRaw _cnUmdUser;			/* in process by user			*/

    /* 
     * upstream messages 
     */
    ChainRaw _cnUmuQueue;			/* sent upstream 			*/
    ChainRaw _cnUmuDriver;			/* in process by driver			*/
    ChainRaw _cnUmuUser;			/* in process by user			*/

    ChainRaw *pcnUmuPool() { return &_cnUmuPool; }
    ChainRaw *pcnUmdPool() { return &_cnUmdPool; }
    ChainRaw *pcnUmdUser() { return &_cnUmdUser; }	

    ChainRaw *pcnUmuDriver() { return &_cnUmuDriver; }	/* allocated messages	*/
    ChainRaw *pcnUmuQueue() { return &_cnUmuQueue; }	/* msgs sent upstream   */
    ChainRaw *pcnUmuUser() { return &_cnUmuUser; }	

    ChainRaw *pcnUmdQueue() { return &_cnUmdQueue; }
    ChainRaw *pcnUmdDriver() { return &_cnUmdDriver; }
   public:
    /* init buffers  */
    void initPool(UmiMsgDownBuf *pUmd) { _cnUmdPool.push(pUmd); }
    void initPool(UmiMsgUpBuf *pUmu) { _cnUmuPool.push(pUmu); }

    /* move all back to the pools and call finiPool()	*/
    void finiQueues();			
    void finiPools();			
    virtual void finiPool(Link *);

    #ifndef _MM_NO_IOSTREAM
     std::ostream &printQueues(std::ostream &os) const ;
    #endif 

    UmiDriver *pud() { return _pud; }
    UmiEnv *pue() { return _pue; }
    int iChan() { return _iChan; }

    /*
     * synchronous open/close
     * only one open()/close() per use. close and reopen
     * the channel for reset.
     */
    virtual int open();
    virtual void close();

    /*
     * pollIn, Out
     */
    int pollIn()  { return pcnUmuQueue()->c(); } 
    UmiMsgUpBuf*getUpBuf();		/* get from up queue, iff pollIn()	*/
    void releaseUpBuf(int iCtx, UmiMsgUpBuf *);	/* most recent got from getUpBuf()	*/

    int pollOut() { return pcnUmdPool()->c(); } 
    UmiMsgDownBuf* allocDownBuf();
    void putDownBuf(int iCtx, UmiMsgDownBuf*);	/* put to driver		*/

    /*
     * send a message upstream, message and buffer must have been obtained
     * by allocUp and are released be this call
     */
    virtual void putUp(int iCtx, Pmu );

    /*
     * allocate an upstream message
     */
    virtual Pmu allocUp(size_t sData);

    /*
     * allocate an upstream message. If the queue is
     * blocked, remove the last message putUp() and reallocate it.
     */
    virtual Pmu allocUpForced(size_t sData);

    /*
     * allocate an upstream message. Declares that the control
     * and data buffers of pmd are no more used and may be
     * reused for the upstream message. 
     */
    virtual Pmu allocUp(size_t sData, Pmd pmd);


    /*
     * release an downstream message, ANY downstream message must
     * be released this way. Clears channel. 
     * Driver::putDown MAY BE CALLED DURING releaseDown() FOR THE CHANNEL.
     */
    virtual void releaseDown(int iCtx, Pmd);

    /* 
     * next timer tick >= now + ti (depends on system clock granularity)
     */
    Time tiNext( const Time &ti ) { return pue()->tiNext(ti); }


    const UmiProto *pup(); 

    void setHandle(void *p) { _phEnv=p; }
    void setDriver(UmiDriver *pud) { _pud=pud; }
    void setChan(int i) { _iChan=i; }
    UmiChannel(UmiDriver *pud, int iChan, UmiEnv *pue);
    virtual ~UmiChannel();
  };

  /*
   * timer interface def
   */
  class UmiCard;
  /*
   * lower message handler (driver) base class 
   */
  class UmiDriver : public Link {
   public:
    typedef int DownE;
   private:
    const char *_szName;
    UmiChannel *_pcn;
    UmiCard *_pca;
    ChainRaw _cnuc;			/* channel list		*/

   protected:
    Time tiNext( const Time &ti ) { return pcn(0)->tiNext(ti); }

   public:
    UmiChannel *pcn(int i) { return static_cast<UmiChannel*>(_cnuc.get(i)); }
    int ccn() { return _cnuc.c(); }

    void startIterChannels( ChainIterator *pit) { _cnuc.strtIter(pit); }
    UmiChannel *iterChannels( ChainIterator *pit) { return static_cast<UmiChannel*>(_cnuc.iter(pit)); }

    UmiCard *pca() const { return _pca; }
    const char *szName() { return _szName ; }		/* driver name, qualified by card's name */



    /* 
     * downstream message put - buffer may be reused for
     * upstream message transfer, see pha()->allocUp().
     * The message channel is locked until pha()->releaseDown() is called.
     */ 
    virtual DownE  putDown(int iCtx, Pmd pmd, int iChan)=0;

    /*
     * per channel protocol descriptor
     */
    virtual const UmiProto *pup(int iChan)=0;

    /*
     * reset any activity related to iChan (open/close processing)
     */
    virtual int open(int iChan);
    virtual void close(int iChan);
    virtual char *print( char *, int );

    /*
     * channel back-enable 
     */
    virtual void backenable(int iCtx, int iChan);

    /* interupt handler */
    virtual int interrupt(int iCtx)=0;

    /* timer handler */
    virtual int timer(int iCtx, const Time &)=0;


    void init(UmiChannel *pcn) { _cnuc.push(pcn); }
    void fini(UmiChannel *pcn) { _cnuc.remove(pcn); }
    UmiDriver(const char *szName, UmiCard *pca);
    virtual ~UmiDriver();
  };

  /*
   * Card abstraction
   */
  class UmiCard {
    ChainRaw _cnud;		/* all drivers		*/
    UmiDriver *_pudCur;		/* "current" driver	*/
   protected:
   public:
    UmiDriver *pudCur() { return _pudCur; }
    int setDriver(UmiDriver *pud) { if (_pudCur) return -1; _pudCur=pud; return 0; }
    int clrDriver(UmiDriver *pud) { if (_pudCur != pud) return -1; _pudCur=0; return 0; }

    void startIterDrivers( ChainIterator *pit) { _cnud.strtIter(pit); }
    UmiDriver *iterDrivers( ChainIterator *pit) { return static_cast<UmiDriver*>(_cnud.iter(pit)); }



    /* per card int and timer */
    virtual int interrupt(int iCtx)=0;
    virtual int timer(int iCtx, const Time &)=0;

    void initDriver(UmiDriver *pud) { _cnud.push(pud); }
    UmiCard():_pudCur(0) {}
    virtual ~UmiCard() {}
  };
}  


#endif


