/*
 * ines Linux UMI driver 
 * 
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/pci.h>
#include <linux/time.h>
#include <linux/spinlock.h>
#include <linux/kdev_t.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <assert.h>
#include "lumi.h"
#include "gpibc/cards.h"

#define PFX "%s: "
//#define SZ_CARD_ID 32


#if 0
# define assert(expr) \
  ((void) ((expr) ? 0 :	printk( KERN_ERR "lumi: " "assertion failed: %s, %s, %d\n", #expr, __FILE__, __LINE__), 0)) \

#endif


  /*
   * tiNext function
   */
  void tiNextTick( struct timespec *ptsNext ) {
    unsigned long j=jiffies+1;		/* REMIND lock	*/
    ptsNext->tv_sec=j/HZ;
    ptsNext->tv_nsec=(j%HZ)*(1000000000/HZ);
  }


  struct Card;
  struct Chan {
    spinlock_t _sl;			/* thread lock mutex		*/
    struct UmiDriver *_pud;
    struct UmiChannel *_puc;
    struct Card *_pca;

    /* wait queues */
    wait_queue_head_t _wqDown;	/* waiting to put message	*/
    wait_queue_head_t _wqUp;	/* waiting to get message	*/
    const char *_szStat;
  };

  #define Chan_isNonBlocking(p) (0)
  #define Chan_lockGet(p) (1)
  #define Chan_lockPut(p) (1)
  #define Chan_unlockGet(p, i) (i)
  #define Chan_unlockPut(p, i) (i)

  /*
   * struct Region 
   */
  struct Region {
    unsigned long _lAddr;
    unsigned long _lLen;
    int _f;
  };

  #define MAX_CHAN 8

  struct Card {
    /* state part */
    char _mzId[SZ_CARD_ID];		/* identifies card physically	*/
    spinlock_t _sl;			/* thread lock mutex		*/
    int _fInit;
    int _fCheck;
    int _fTimerStop;			/* timer shutdown flag		*/
    struct timer_list _timer;		/* timer tick			*/

    const struct CardType *_pct;	/* type of the card		*/
    struct LumiMod *_plm;

    /* card physical resource */
    struct Region _mprg[MAX_PCI_REGION];
    int _iIoRegion;			/* region for controller IO	*/
    int _iMemRegion;			/* region for memory access 	*/
    int _irq;
    int _fIrq;				/* shared/non shared		*/

    struct UmiMemInfo _mpmi[MAX_UMI_MEM];  /* memory for the UMI card driver */
    int _cmi;

    struct UmiCard *_puc;

    struct Chan _mpcn[MAX_CHAN];
    int _ccn;
  };

  #define Card_isInit(p) ((p)->_fInit != 0)
  #define Card_setInit(p) ((p)->_fInit = 1)
  #define Card_clrInit(p) ((p)->_fInit = 0)

  int Card_isInit_locked( struct Card *pca ) {
    /*
     * check if the card is available. This is concurrent
     * to the hotplug probing process.
     */
    int i,f;
    spin_lock_irqsave(&pca->_sl, f);
    i=Card_isInit(pca);
    spin_unlock_irqrestore(&pca->_sl, f);
    return i;
  }

  #define Chan_stat(pcn, sz) (pcn)->_szStat=(sz)

  int Chan_canput(struct Chan *pcn) { 
    int r;
    int f;
    spin_lock_irqsave(pcn->_pca->_sl, f);
    r=UmiChannel_canput(pcn->_puc); 
    spin_unlock_irqrestore(pcn->_pca->_sl, f);
    return r;
  }

  int Chan_canget(struct Chan *pcn) { 
    int r;
    int f;
    spin_lock_irqsave(pcn->_pca->_sl, f);
    r=UmiChannel_canget(pcn->_puc); 
    spin_unlock_irqrestore(pcn->_pca->_sl, f);
    return r;
  }



  /*
   * Card structure allocation/release
   */
  struct Card *findCard(struct LumiMod *plm, const char *sz) {
    int i;
    for (i=0; i<plm->_cMaxCard; i++) if (!strcmp(sz,plm->_pca[i]._mzId)) return plm->_pca+i;
    return 0;
  }
  struct Card *allocCard(struct LumiMod *plm, const char *sz) {
    int i;
    for (i=0; i<plm->_cMaxCard; i++) {
      struct Card *pca=plm->_pca+i;
      if (!strcmp("",pca->_mzId)) {
        strcpy(pca->_mzId, sz);
	pca->_plm=plm;
        return pca;
      }
    }
    return 0;
  }
  void freeCard(struct LumiMod*plm, struct Card *pca) {
    memset(pca->_mzId, 0, sizeof(pca->_mzId));
  }

  int Card_num(struct LumiMod*plm, struct Card *pca) { return pca - plm->_pca; }


  /*
   * debug messages
   */
  void Chan_debug(struct LumiMod *plm, struct Chan *pcn, char *szMsg, int iStat) {
    if (*plm->_pfDebug) printk(KERN_DEBUG PFX " %s:%s:%d %s %x\n", plm->_szDid, pcn->_pca->_mzId,  UmiDriver_szName(pcn->_pud), UmiChannel_num(pcn->_puc), szMsg, iStat);
  }
  void __assert_fail (__const char *__assertion, __const char *__file, unsigned int __line, __const char *__function) {
    panic("lumi %s %s %d %s\n", __assertion, __file, __line, __function);
    //printk(KERN_EMERG "lumi: %s %s %d %s\n", __assertion, __file, __line, __function);
  }


  /*
   * kmalloc all memory required
   */
  int kmalloc_all( struct UmiMemInfo *pmi, int c) {
    int i;
    for (i=0; i<c; i++, pmi++) {
      pmi->_p=kmalloc(pmi->_s, GFP_KERNEL);
      if (!pmi->_p) return -ENOMEM;
    }
    return 0;
  }

  int kfree_all( struct UmiMemInfo *pmi, int c) {
    int i;
    for (i=0; i<c; i++, pmi++) if (pmi->_p) kfree(pmi->_p);
    return 0;
  }


  /* per card timer */
  void lumitimer(unsigned long p) {
    struct Card *pca=(struct Card*)p;
    int f;
    //char mz[80];

    spin_lock_irqsave(pca->_sl, f);	
    //debugmsg(pif, "timer", GpibInterface_print(pif->_pgi, mz, sizeof(mz)));
    if (1) {
      struct timespec ts;
      unsigned long j=jiffies;		/* REMIND lock		*/
      ts.tv_sec=j/HZ;
      ts.tv_nsec=(j%HZ)*(1000000000/HZ);

      UmiCard_timer(pca->_puc, 333, &ts);
    } else {
      printk(KERN_DEBUG PFX "timer call  SKIPPED\n", pca->_plm->_szDid);
    }
    spin_unlock_irqrestore(pca->_sl, f);	
     

    /* reschedule timer */
    pca->_timer.expires=jiffies + (HZ/pca->_plm->_pum->_tps) ;
    if (!pca->_fTimerStop) add_timer(&pca->_timer);
  }

  /*
   * per card interrupt
   */
  void lumiintr(int irq ,void * dev_id, struct pt_regs *regs) {
    struct Card *pca=(struct Card*)dev_id;


    /* run the interface interrupt procedure */
    int f;
    spin_lock_irqsave(pca->_sl, f);	

    /* _all_ other (timer) local interrupts are off now */

    if (pca->_fCheck) panic("lumi interrupt recursion\n");
    pca->_fCheck=1;
    
    assert(pca->_puc);
    UmiCard_interrupt(pca->_puc, 666);
    pca->_fCheck=0;
    spin_unlock_irqrestore(pca->_sl, f);	
  } 


  static int lumi_read_procmem( char* pchBuf, char **ppch, off_t offs , int c, int *piEof, void *p) {
    struct LumiMod *plm=(struct LumiMod*) p;

    int i;
    char *pch=pchBuf;
    char *pchLim=pchBuf+c-80;

    for (i=0; i<plm->_cMaxCard && pch<pchLim; i++) {
      struct Card *pca=plm->_pca+i;
      if (Card_isInit(pca)) {
	pch += sprintf(pch, "card %d: ", i); 
        const struct CardType *pct=pca->_pct;
        pch += sprintf(pch, "id=%s name=%s arch=%s bus=%c ", pca->_mzId, pct->_szName, pct->_szArch, pct->_eBus);
        pch += sprintf(pch, "io=%lx irq=%d", pca->_mprg[pca->_iIoRegion]._lAddr , pca->_irq);
        if (pca->_iMemRegion != -1) pch += sprintf(pch, " mem=%lx", pca->_mprg[pca->_iMemRegion]._lAddr);
	pch += sprintf(pch, "\n"); 

	{
	  struct Chan *pcn=pca->_mpcn;
	  for (; pcn < pca->_mpcn+pca->_ccn; pcn++) {
	    char mpch[80];
	    UmiDriver_print(pcn->_pud, mpch, 80);
	    pch += sprintf(pch, " %s:%d getok:%d putok:%d stat=%s core=%s\n", UmiDriver_szName(pcn->_pud), UmiChannel_num(pcn->_puc), Chan_canget(pcn), Chan_canput(pcn), pcn->_szStat ? pcn->_szStat : "nil", mpch  );
	  }
	}
	#if 0
	{
	  struct UmiDriver *pud;
	  struct UmiIter it;
	  UmiCard_startIterDrivers(pca->_puc, &it);
	  while ((pud=UmiCard_iterateDrivers(pca->_puc, &it))) {
	    struct UmiIter it;
	    struct UmiChannel *puc;

	    pch += sprintf(pch, " %s\n", UmiDriver_szName(pud));
	    UmiDriver_startIterChannels(pud, &it);
	    while ((puc=UmiDriver_iterateChannels(pud, &it))) {
	      pch += sprintf(pch, "  %d\n", UmiChannel_num(puc));
	    }
          }
        }
	#endif

      }
    }

    *piEof=1;
    return pch-pchBuf;
  }


  /*
   * signal procs called from UMI channel
   */
  void lumicanget( int iCtx , void *p) {
    assert(p);
    struct Chan *pcn=(struct Chan *)p;
    wake_up_interruptible(&pcn->_wqUp);
  }

  void lumicanput( int iCtx , void *p) {
    assert(p);
    struct Chan *pcn=(struct Chan *)p;
    wake_up_interruptible(&pcn->_wqDown);
  }

  struct UmiEnvC ue = { tiNextTick, lumicanget, lumicanput } ;

  int lumi_ioctl(struct LumiMod*plm, struct inode * pin,struct file *fp, unsigned int cmd, unsigned long arg ) {
    struct Chan *pcn;
    pcn=(struct Chan*)(fp->private_data);
    assert(pcn);
    Chan_debug(plm, pcn, "ioctl", cmd);

    /*
     * this thread now owns pcn and pcn->_puc
     */
    if (cmd & GIOC_PUTMSG) {
      int iErr;

      /*
       * serialize this among concurrent threads
       */
      if (Chan_lockPut(pcn)) {

	Chan_stat(pcn, "ioctl-put");

	/* put a message to the interface */
	if (Chan_isNonBlocking(pcn) && !Chan_canput(pcn)) return Chan_unlockPut(pcn, -EAGAIN);

	Chan_debug(plm, pcn, "sleep", cmd);
	Chan_stat(pcn, Chan_canput(pcn) ? "ioctl-put-noblock" : "ioctl-put-block");

	/* wait until we can put */
	iErr = wait_event_interruptible(pcn->_wqDown, (Chan_canput(pcn)));
	if (iErr) {
	  Chan_debug(plm, pcn, "interrupted", iErr);
	  return Chan_unlockPut(pcn, iErr);
	}

	Chan_stat(pcn, "ioctl-putok");
	Chan_debug(plm, pcn, "wakeup", cmd);


	struct UmiMsgDownBuf *pdb;
	{
	  int f;
	  spin_lock_irqsave(pcn->_pca->_sl, f);
	  pdb=UmiChannel_allocDownBuf(pcn->_puc);
	  spin_unlock_irqrestore(pcn->_pca->_sl, f);
	}

	Chan_stat(pcn, "ioctl-put-alloc");

	/* verify and copy in the control part */
	struct UmiControlStruct *pucU=(void*)arg;
	size_t sizeC;
	if ((get_user(sizeC, &pucU->_size))) return Chan_unlockPut(pcn, -EFAULT);

	if (!access_ok(VERIFY_READ, (void*)arg, sizeC) || sizeC > SZ_CTRL_MAX ) return Chan_unlockPut(pcn, -EFAULT);
	if (__copy_from_user(UmiMsgDownBuf_pControl(pdb), (void*)arg, sizeC)) return Chan_unlockPut(pcn, -EFAULT);

	struct UmiControlStruct *puc=UmiMsgDownBuf_pControl(pdb);
	struct UmiDataStruct *pud=&puc->_ud;
	size_t sUserData=pud->_pchLimBuf-pud->_pchBase;
	if (!access_ok(VERIFY_READ, pud->_pchBase, sUserData)) return Chan_unlockPut(pcn, -EFAULT);
	if (sUserData > UmiMsgDownBuf_sizeDataMax(pdb)) return Chan_unlockPut(pcn, -EFAULT);
	if (__copy_from_user(UmiMsgDownBuf_pData(pdb), pud->_pchBase, sUserData)) return Chan_unlockPut(pcn, -EFAULT);
	UmiMsgDownBuf_setDataLim(pdb, sUserData);

	Chan_stat(pcn, "ioctl-put-copied");
	/* 
	 * send the message downstream 
	 */
	{
	  int f;

	  spin_lock_irqsave(pcn->_pca->_sl, f);

	  Chan_debug(plm, pcn, "put", 0);
	  UmiChannel_putDownBuf(pcn->_puc, 0, pdb);

	  spin_unlock_irqrestore(pcn->_pca->_sl, f);
	}
	Chan_stat(pcn, "ioctl-put-done");
	(void)Chan_unlockPut(pcn, 0);
      }
      else return -EAGAIN;
    }

    if (cmd & GIOC_GETMSG) {
	int iErr;

      if (Chan_lockGet(pcn)) {

	Chan_stat(pcn, "ioctl-get");
	/* get a message from the interface */
	if (Chan_isNonBlocking(pcn) && !Chan_canget(pcn)) return Chan_unlockGet(pcn, -EAGAIN);

	iErr = wait_event_interruptible(pcn->_wqUp, Chan_canget(pcn));
	if (iErr) {
	  Chan_stat(pcn, "ioctl-get-intr");
	  return Chan_unlockGet(pcn,iErr);
        }

	Chan_stat(pcn, "ioctl-get-ok");

	struct UmiMsgUpBuf *pub;

	/* get the message */
	{
	  int f;
	  spin_lock_irqsave(pcn->_pca->_sl, f);

	  Chan_debug(plm, pcn, "get", 0);
	  pub=UmiChannel_getUpBuf(pcn->_puc);

	  spin_unlock_irqrestore(pcn->_pca->_sl, f);
	}
	Chan_stat(pcn, "ioctl-get-buf");


	/*
	 * save the pointers to the user data buffer
	 * REMIND: How to return (set) the user data limit ?
	 */
	struct UmiControlStruct *pucU=(void*)arg;
	struct UmiDataStruct *pudU=&pucU->_ud;
	struct UmiDataStruct ud;
	if ((get_user(ud._pchBase, &pudU->_pchBase)) || (get_user(ud._pchLimBuf, &pudU->_pchLimBuf))) return Chan_unlockGet(pcn, -EFAULT);


	/* copy control part to user buffer */
	if (!access_ok(VERIFY_WRITE, (void*)arg, UmiMsgUpBuf_sizeControl(pub))) return Chan_unlockGet(pcn, -EFAULT);
	if (__copy_to_user((void*)arg, UmiMsgUpBuf_pControl(pub), UmiMsgUpBuf_sizeControl(pub))) return Chan_unlockGet(pcn, -EFAULT);

	/* copy data part to user buffer */
	if (UmiMsgUpBuf_sizeData(pub) > (size_t)(ud._pchLimBuf-ud._pchBase)) return Chan_unlockGet(pcn, -EFAULT);
	if (!access_ok(VERIFY_WRITE, ud._pchBase, ud._pchLimBuf-ud._pchBase)) return Chan_unlockGet(pcn, -EFAULT);
	if (__copy_to_user(ud._pchBase, UmiMsgUpBuf_pData(pub), UmiMsgUpBuf_sizeData(pub))) return Chan_unlockGet(pcn, -EFAULT);

	/* (reset) set result buffer data */
	ud._pchLimDat=ud._pchBase+UmiMsgUpBuf_sizeData(pub);
	if (!access_ok(VERIFY_WRITE, pudU, sizeof(struct UmiDataStruct))) return Chan_unlockGet(pcn, -EFAULT);
	if (__copy_to_user(pudU, &ud, (sizeof(struct UmiDataStruct)))) return Chan_unlockGet(pcn, -EFAULT);

	Chan_stat(pcn, "ioctl-get-copied");
	/* release the buffer */
	{
	  int f;
	  spin_lock_irqsave(pcn->_pca->_sl, f);

	  Chan_debug(plm, pcn, "release", 0);
	  UmiChannel_releaseUpBuf(pcn->_puc, 0, pub);

	  spin_unlock_irqrestore(pcn->_pca->_sl, f);
	}

	Chan_stat(pcn, "ioctl-get-done");
        (void)Chan_unlockGet(pcn,0);
      }
      Chan_stat(pcn, 0);
    }
    return 0;
  }

  int lumi_open(struct LumiMod*plm, struct inode * pin, struct file *fp) {
    int iChan;
    int iCard;
    struct Card *pca;
    struct Chan *pcn;

    iCard=GETCARD(MINOR(pin->i_rdev));
    if (iCard < 0 || iCard >= plm->_cMaxCard) return -ENODEV;
    pca=plm->_pca+iCard;

    /*
     * check if the card is available. This is concurrent
     * to the hotplug probing process.
     */
    if (!Card_isInit_locked(pca)) return -ENODEV;

    iChan=GETCHAN(MINOR(pin->i_rdev));
    if (iChan < 0 || iChan >= pca->_ccn) return -ENODEV;


    /*
     * open the UMI channel
     */
    pcn=pca->_mpcn+iChan;

    if (UmiChannel_open(pcn->_puc)) return -EIO;
    fp->private_data=pcn;

    Chan_debug(plm, pcn, "open", pin->i_rdev);
    return 0;
  }
  int lumi_release(struct LumiMod*plm, struct inode * pin, struct file *fp) {
    struct Chan *pcn=((struct Chan*)(fp->private_data));
    Chan_debug(plm, pcn, "close", pin->i_rdev);
    UmiChannel_close(pcn->_puc);
    return 0;
  }

  int lumi_mmap(struct LumiMod *plm, struct file*pFil, struct vm_area_struct *pVma) {
    const struct Card *pca=((struct Chan*)(pFil->private_data))->_pca;
    if (pca->_iMemRegion == -1) return -ENODEV;

    unsigned long lOff=pVma->vm_pgoff << PAGE_SHIFT;
    unsigned long lPhys=pca->_mprg[pca->_iMemRegion]._lAddr + lOff;
    unsigned long lSizeV=pVma->vm_end-pVma->vm_start;
    unsigned long lSizeP=pca->_mprg[pca->_iMemRegion]._lLen - lOff;

    printk(KERN_DEBUG PFX "mmap (%s) off=%lx phys=%lx vsize=%lx psize=%lx\n", plm->_szDid, pca->_mzId, lOff, lPhys, lSizeV, lSizeP);

    if (lSizeV > lSizeP) return -EINVAL;
    if (remap_page_range(pVma->vm_start, lPhys, lSizeV, pVma->vm_page_prot)) return -EAGAIN;
    return 0;
  }

  /*
   * Card f/initialization
   */
  int Card_init( struct Card *pca) { 
    int i;
    const char *szDid=pca->_plm->_szDid;

    assert(pca->_pct);
    /*
     * init UMI
     */
    pca->_cmi=UmiCard_getMem(pca->_plm->_pum, pca->_mpmi, MAX_UMI_MEM);
    assert(pca->_cmi >= 0);
    i=kmalloc_all(pca->_mpmi, pca->_cmi);
    if (i) {
      printk(KERN_ERR PFX "can't kmalloc() %d blocks\n", szDid, pca->_cmi);
      return i;
    }
    else {
      printk(KERN_DEBUG PFX "done kmalloc() %d blocks\n", szDid, pca->_cmi);
    }

    /*
     * create UMI card, driver, and channels
     */
    pca->_puc=UmiCard_new(pca->_plm->_pum, &ue, pca->_mpmi, pca->_cmi, pca->_pct, pca->_mprg[pca->_iIoRegion]._lAddr);
    printk(KERN_DEBUG PFX "umi: new card %p\n", szDid, pca->_puc);

    {
      struct Chan *pcn=pca->_mpcn;

      struct UmiDriver *pud;
      struct UmiIter it;
      UmiCard_startIterDrivers(pca->_puc, &it);
      while ((pud=UmiCard_iterateDrivers(pca->_puc, &it))) {
	struct UmiIter it;
	struct UmiChannel *puc;

	UmiDriver_startIterChannels(pud, &it);
	while ((puc=UmiDriver_iterateChannels(pud, &it))) {
	  if (pcn < pca->_mpcn+MAX_CHAN) {
	    /* init channel */
	    pcn->_pud=pud;
	    pcn->_puc=puc;
	    pcn->_pca=pca;
	    UmiChannel_setHandle(pcn->_puc, pcn);		/* for callback	*/

	    init_waitqueue_head(&pcn->_wqUp);
	    init_waitqueue_head(&pcn->_wqDown);

	    printk(KERN_DEBUG PFX "umi: new chan %d %s %d\n", szDid, pcn-pca->_mpcn, UmiDriver_szName(pcn->_pud), UmiChannel_num(pcn->_puc));
	    pcn++;
	    pca->_ccn++;
	  }
	}
      }
    }

    /*
     * install timer tick
     */
    /* init the timer */
    pca->_fTimerStop=0;
    init_timer(&pca->_timer);
    pca->_timer.function=lumitimer;
    pca->_timer.data=(unsigned long)pca;
    pca->_timer.expires=jiffies + (HZ/pca->_plm->_pum->_tps) ;
    add_timer(&pca->_timer);

    /* 
     * install interrupt
     */
    if( request_irq(pca->_irq, lumiintr, pca->_fIrq, szDid, pca) ) {
      printk(KERN_ERR PFX "IRQ %d not available\n", szDid, pca->_irq);

      /*
       * REMIND cleanup
       */
      //GpibInterface_delete(pgi);
      return -EAGAIN;
    }

    Card_setInit(pca);
    return 0;
  }

  void Card_fini(struct Card *pca) {
    Card_clrInit(pca);

    pca->_fTimerStop=1;
    del_timer_sync(&pca->_timer); 	/* FIXME: notify concurrent ieeetimer not to add the timer again */
    if (pca->_irq) free_irq(pca->_irq, pca);

    /* REMIND: dtor call */

    kfree_all(pca->_mpmi, pca->_cmi);
  }

  /*
   * PCI probe/remove
   */
  static char * szIdentPci( char *mpchBuf, struct pci_dev *pdv ) {
    sprintf(mpchBuf, "pci:%s", pdv->slot_name);
    return mpchBuf;
  }

  static char * szIdentIsa( char *mpchBuf, int io ) {
    sprintf(mpchBuf, "isa:%x", io);
    return mpchBuf;
  }

  /*
   * PCI card f/initialization
   */
  int Card_initPci( struct Card *pca, struct pci_dev *pdv, const struct pci_device_id *pdi, const char *szIdent) {
    int i;
    const struct CardType *pct;
    const char *szDid=pca->_plm->_szDid;


    if (pci_enable_device(pdv)) return -EIO;

    assert(pdi->driver_data >= 0 && pdi->driver_data < CNT_CARDTYPES);
    pct=mpct+pdi->driver_data;
    pca->_pct=pct;
    pca->_iIoRegion=-1;
    pca->_iMemRegion=-1;

    printk(KERN_DEBUG PFX "card type=%s arch=%s\n",szDid, pct->_szName, pct->_szArch);

    /*
     * get regions 
     */

    for(i=0; i<MAX_PCI_REGION; i++) {
      unsigned long lAddr, lLen, lFlags;
      const struct RegionDesc *prd=pct->_mprd+i;
      struct Region *prg=pca->_mprg+i;

      if (prd->_iBar == -1) break;

      lAddr=pci_resource_start(pdv, prd->_iBar);
      lLen=pci_resource_len(pdv, prd->_iBar);
      lFlags=pci_resource_flags(pdv, prd->_iBar);

      if (!lAddr) {
	printk(KERN_ERR PFX "PCI region #%d not available\n", szDid, prd->_iBar);
	return -ENODEV;
      }
      printk(KERN_DEBUG PFX "PCI Region %d addr=%lx len=%lx flags=%lx\n", szDid, prd->_iBar, lAddr, lLen, lFlags);

      if (RG_ISMEM(prd->_f)) {
        /* MEM space */
	if (check_mem_region(lAddr,lLen)) {
	  printk(KERN_ERR PFX "memory region #%d not available\n", szDid, prd->_iBar);
	  return -ENODEV;
	}
	request_mem_region(lAddr,lLen, szDid);
	prg->_lAddr=lAddr;
	prg->_lLen=lLen;
	prg->_f=lFlags;
	pca->_iMemRegion=i;
      }
      else {
        /* IO space */
	if (check_region(lAddr,lLen)) {
	  printk(KERN_ERR PFX "io region #%d not available\n", szDid, prd->_iBar);
	  return -ENODEV;
	}

	request_region(lAddr,lLen, szDid);
	prg->_lAddr=lAddr;
	prg->_lLen=lLen;
	prg->_f=lFlags;

	if(prd->_f & RG_AMCC5920_CONF) { 
	  outb(0x87, lAddr+0x60); 		// max(7)  waitstates for pass through configuration
	  outb(0x2c, lAddr+0x38+1); 	// set bit 13 in INTCSR to enable interrupts
	  printk(KERN_DEBUG PFX "AMCC5920 configured %lx\n", szDid, lAddr);
	} 
	else if (prd->_f & RG_IOBASE) { 
	  pca->_iIoRegion=i;
	} 
      }
    }
    assert(pca->_iIoRegion >= 0);	/* required */

    /* request irq */

    
    pca->_irq=pdv->irq;
    pca->_fIrq=SA_SHIRQ;
    printk(KERN_DEBUG PFX "IRQ %d\n", szDid, pca->_irq);

    return 0;
  }

  void Card_finiPci(struct Card *pca) {
    int i;

    for(i=0; i<MAX_PCI_REGION; i++) {
      struct Region *prg=pca->_mprg+i;

      if (prg->_f & IORESOURCE_IO) release_region(prg->_lAddr,prg->_lLen);
      else if (prg->_f & IORESOURCE_MEM) release_mem_region(prg->_lAddr,prg->_lLen);
    }
  }

  int Card_initIsa(struct Card *pca, int io, int irq, const char *szIdent) {
    const struct CardType *pct=(mpct+CT_GPIB_PC);
    const char *szDid=pca->_plm->_szDid;


    pca->_pct=pct;
    pca->_iIoRegion=0;
    pca->_iMemRegion=-1;
    struct Region *prg=pca->_mprg+0;


    printk(KERN_DEBUG PFX "try IO region %x(%d)\n", szDid, io, pct->_isaLen);
    if (check_region(io,pct->_isaLen)) {
      printk(KERN_ERR PFX "IO region %x(%d) not available\n", szDid, io, pct->_isaLen);
      return -ENODEV;
    }

    request_region(io,pct->_isaLen, szDid);
    prg->_lAddr=io;
    prg->_lLen=pct->_isaLen;
    prg->_f=IORESOURCE_IO;

    printk(KERN_DEBUG PFX "card type=%s arch=%s\n",szDid, pct->_szName, pct->_szArch);


    /* request irq */
    pca->_irq=irq;
    pca->_fIrq=0;
    printk(KERN_DEBUG PFX "IRQ %d\n", szDid, pca->_irq);


    Card_setInit(pca);
    return 0;
  }

  void Card_finiIsa(struct Card *pca) {
    int i;

    for(i=0; i<MAX_PCI_REGION; i++) {
      struct Region *prg=pca->_mprg+i;

      if (prg->_f & IORESOURCE_IO) release_region(prg->_lAddr,prg->_lLen);
      else if (prg->_f & IORESOURCE_MEM) release_mem_region(prg->_lAddr,prg->_lLen);
    }
  }


  /*
   * probe() might be executed by indepent threads.
   * Thus the access to the card structure is locked.
   */
  int lumi_probe(struct LumiMod *plm, struct pci_dev *pdv, const struct pci_device_id *pdi) {
    char mpch[SZ_CARD_ID];
    struct Card *pca;

    printk(KERN_INFO PFX "pci probe (%s)\n", plm->_szDid, pdv->name);

    if ((pca=findCard(plm, szIdentPci(mpch, pdv)))!=NULL) {
      /* already registered */
      printk(KERN_INFO PFX "already initialized %s@%d\n", plm->_szDid, mpch,Card_num(plm,pca));
      return -1;
    }

    /* find a free slot for the card */
    if (!(pca=allocCard(plm, mpch))) {
      printk(KERN_INFO PFX "cannot allocate card %s\n", plm->_szDid, mpch);
      return -1;
    }

    printk(KERN_INFO PFX "initialize card [%d] %s\n",plm->_szDid, Card_num(plm,pca), mpch);

    /* initialize */
    if ( Card_initPci(pca, pdv, pdi, mpch)) return -1;
    else pci_set_drvdata(pdv, pca);

    return (Card_init(pca));
  }

  void lumi_remove(struct LumiMod *plm, struct pci_dev *pdv) {
    struct Card *pca=pci_get_drvdata(pdv);
    if (pca == NULL) return;



    printk(KERN_INFO PFX "remove card %s@%d\n",plm->_szDid, pca->_mzId, Card_num(plm,pca));

    Card_fini(pca);
    Card_finiPci(pca);
    freeCard(plm, pca);
  }

  int enum_isa_devices(struct LumiMod *plm) {
    int i;
		  
    /* Check for user specified ISA devices */
    for (i=0 ;(i < plm->_cMaxIsa) && plm->_piIoIsa[i] && plm->_piIrqIsa[i]; i++) {
      struct Card *pca;
      char mpch[SZ_CARD_ID];

      if (*plm->_pfDebug) printk(KERN_DEBUG PFX "ISA device specified io=%04X,irq=%d\n", plm->_szDid, plm->_piIoIsa[i], plm->_piIrqIsa[i] ); 

      if ((pca=findCard(plm, szIdentIsa(mpch, plm->_piIoIsa[i])))!=NULL) {
	/* already registered */
	printk(KERN_INFO PFX "already initialized %s@%d\n",plm->_szDid, mpch,Card_num(plm, pca));
	return -1;
      }

      /* find a free slot for the card */
      if (!(pca=allocCard(plm, mpch))) {
	printk(KERN_INFO PFX "cannot allocate card %s\n",plm->_szDid, mpch);
	return -1;
      }
      printk(KERN_INFO PFX "initialize card [%d, %d] %s\n",plm->_szDid, i, Card_num(plm, pca), mpch);

      /* initialize */
      if ( Card_initIsa(pca, plm->_piIoIsa[i], plm->_piIrqIsa[i], mpch) || Card_init(pca) ) return -1;
    }
    return 0;
  }

  int free_isa_devices(struct LumiMod *plm) {
    int i;
		  
    /* Check for user specified ISA devices */
    for (i=0 ;(i < plm->_cMaxIsa) && plm->_piIoIsa[i] && plm->_piIrqIsa[i]; i++) {
      struct Card *pca;
      char mpch[SZ_CARD_ID];


      if ((pca=findCard(plm, szIdentIsa(mpch, plm->_piIoIsa[i])))==NULL) {
	/* error: released */
	printk(KERN_ERR PFX "already released %s@%d\n",plm->_szDid, mpch,Card_num(plm, pca));
	return -1;
      }
      printk(KERN_INFO PFX "release isa card %s@%d\n", plm->_szDid, pca->_mzId, Card_num(plm,pca));

      Card_fini(pca);
      Card_finiIsa(pca);
      freeCard(plm, pca);
    }
    return 0;
  }



  /*
   * init/cleanup module
   */
  void lumi_cleanup_module(struct LumiMod *plm) {
    devfs_unregister(plm->_hDfs);
    devfs_unregister_chrdev(plm->_iMajor,plm->_szDid);
    pci_unregister_driver (plm->_pPdr);
    free_isa_devices(plm);

    remove_proc_entry(plm->_szProcDir, 0);
    if (plm->_pca) kfree(plm->_pca);
  }

  int lumi_init_module(struct LumiMod *plm) {
    printk(KERN_INFO "%s: %s driver %s for Linux\n", plm->_szDid, plm->_szName,  plm->_szDvs);

    plm->_pca=(struct Card*)kmalloc(sizeof(struct Card) * plm->_cMaxCard, GFP_KERNEL);

    if (!plm->_pca) {
      printk(KERN_ERR "%s: can't alloc card memory\n", plm->_szDid);
      return -ENOMEM;
    }
    else {
      struct Card *pca;
      for (pca=plm->_pca; pca< plm->_pca+plm->_cMaxCard; pca++) {
	memset(pca, 0, sizeof(struct Card));
        spin_lock_init(pca->_sl);
      }
    }

  if (enum_isa_devices(plm)) return -ENODEV;

    /* register for pci */
    (void) pci_register_driver(plm->_pPdr);

    plm->_iMajor=0;
    plm->_iMajor=devfs_register_chrdev(plm->_iMajor,plm->_szDid,plm->_pFops);
    if(plm->_iMajor<0) {
      printk(KERN_ERR "%s: can't get major %d\n",plm->_szDid, plm->_iMajor);
      lumi_cleanup_module(plm);
      return -ENODEV;
    }
    plm->_hDfs=devfs_mk_dir(NULL, plm->_szDid, NULL);

    if (!create_proc_read_entry(plm->_szProcDir, 0, 0, lumi_read_procmem, plm)) {
      printk(KERN_ERR "%s: can't create <proc>/%s\n", plm->_szDid, plm->_szProcDir);
      return -ENOMEM;
    }

    return 0;	/* keep driver for hotplug events */
  }


